Package: argparse
Type: Package
Title: Command Line Optional and Positional Argument Parser
Version: 2.3.1
Authors@R: c(person("Trevor L.", "Davis", role=c("aut", "cre"),
             email="trevor.l.davis@gmail.com",
             comment = c(ORCID = "0000-0001-6341-4639")),
    person("Allen", "Day", role="ctb", comment="Some documentation and examples ported from the getopt package."),
    person("Python Software Foundation", role="ctb", comment="Some documentation from the argparse Python module."),
    person("Paul", "Newell", role="ctb"))
Description: A command line parser to
    be used with 'Rscript' to write "#!" shebang scripts that gracefully
    accept positional and optional arguments and automatically generate usage.
License: GPL (>= 2)
Copyright: See file (inst/)COPYRIGHTS.
URL: https://github.com/trevorld/r-argparse,
        https://trevorldavis.com/R/argparse/
BugReports: https://github.com/trevorld/r-argparse/issues
LazyLoad: yes
Depends: R (>= 3.6.0)
Imports: findpython, jsonlite, R6
SystemRequirements: python (>= 3.2)
Suggests: knitr (>= 1.15.19), rmarkdown, testthat
VignetteBuilder: knitr
RoxygenNote: 7.3.3
Encoding: UTF-8
NeedsCompilation: no
Packaged: 2025-10-08 20:05:05 UTC; trevor
Author: Trevor L. Davis [aut, cre] (ORCID:
    <https://orcid.org/0000-0001-6341-4639>),
  Allen Day [ctb] (Some documentation and examples ported from the getopt
    package.),
  Python Software Foundation [ctb] (Some documentation from the argparse
    Python module.),
  Paul Newell [ctb]
Maintainer: Trevor L. Davis <trevor.l.davis@gmail.com>
Repository: CRAN
Date/Publication: 2025-10-08 21:20:02 UTC
